from PySide6.QtCore import Qt
from PySide6.QtWidgets import (
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QSplitter,
    QTabWidget,
    QWidget,
)

from bn_modeller.models import FilterPairTableSQLProxyModel, PairTableSQLProxyModel
from bn_modeller.models.feature_sqltable_model import (
    FeatureSqlTableModel,
    PersistanceCheckableFeatureListProxyModel,
)
from bn_modeller.widgets import DependencySetupTableView, SelectableListView
from bn_modeller.widgets.bn_visualization_view import BayesianNetView
from bn_modeller.widgets.vertical_label import QVertivalLabel


class BayesianNetworkPageWidget(QWidget):
    def __init__(self, parent: QWidget | None = None, f=Qt.WindowType()):
        super().__init__(parent, f)
        self._init_ui()

    def _init_ui(self):
        self.mainLayout = QHBoxLayout(self)

        self.tabWidget = QTabWidget()

        # Dependency tab

        # Feature selection
        self.dependencyTabWidget = QSplitter()
        self.featureSelectorView = SelectableListView()
        self.dependencyTabWidget.addWidget(self.featureSelectorView)
        self.dependencyTabWidget.setStretchFactor(0, 1)

        # Dependency Table
        depTableWidget = QWidget()
        depTableLayout = QGridLayout()

        self._depTable = DependencySetupTableView()
        depTableLayout.addWidget(self._depTable, 1, 1)

        dependentLabel = QVertivalLabel("Independent")
        depTableLayout.addWidget(dependentLabel, 1, 0, Qt.AlignmentFlag.AlignCenter)

        dependentLabel = QLabel("Dependent")
        depTableLayout.addWidget(dependentLabel, 0, 1, Qt.AlignmentFlag.AlignCenter)

        depTableWidget.setLayout(depTableLayout)
        self.dependencyTabWidget.addWidget(depTableWidget)
        self.dependencyTabWidget.setStretchFactor(1, 2)

        self.tabWidget.addTab(self.dependencyTabWidget, self.tr("Dependency"))

        # Visualization Tab
        self.visualizationTabWidget = BayesianNetView()

        self.tabWidget.addTab(self.visualizationTabWidget, self.tr("Visulization"))

        # Finalization
        self.mainLayout.addWidget(self.tabWidget)
        self.setLayout(self.mainLayout)

    def setModels(self, pairTableSQLProxyModel: PairTableSQLProxyModel):
        # TODO: т.к. при реализации обсчета байесовских сетей нужно использовать
        # self._pairTableSQLProxyModel, вероятно, её нужно будет вытащить наружу
        # или сделать рассчеты дочерним объектом этой страницы

        self._featureCheckableSortFilterProxyModel = (
            PersistanceCheckableFeatureListProxyModel()
        )
        self._featureCheckableSortFilterProxyModel.setSourceModel(
            pairTableSQLProxyModel.getFeatureSqlTableModel()
        )

        self.featureSelectorView.setModel(self._featureCheckableSortFilterProxyModel)
        self.featureSelectorView.setModelColumn(
            pairTableSQLProxyModel.getFeatureSqlTableModel().fieldIndex(
                pairTableSQLProxyModel.getFeatureSqlTableModel().column_name
            )
        )

        self._pairTableSQLProxyModel = FilterPairTableSQLProxyModel()
        self._pairTableSQLProxyModel.setSourceModel(pairTableSQLProxyModel)
        self._pairTableSQLProxyModel.setFilterModel(
            self._featureCheckableSortFilterProxyModel,
            pairTableSQLProxyModel.getFeatureSqlTableModel().fieldIndex(
                FeatureSqlTableModel.column_id
            ),
        )
        self._depTable.setModel(self._pairTableSQLProxyModel)

        self.visualizationTabWidget.setModels(self._pairTableSQLProxyModel)



from PySide6.QtCore import QAbstractItemModel, Qt, Signal, Slot
from PySide6.QtWidgets import QAbstractItemView, QHBoxLayout, QListView, QWidget

from bn_modeller.models.sample_sqltable_model import SampleSqlTableModel
from bn_modeller.widgets.all_samples_view import AllSamplesView


class SelectableListView(QListView):
    def __init__(self, parent: QWidget | None = None):
        super().__init__(parent)
        self.setSelectionMode(QAbstractItemView.SelectionMode.MultiSelection)
        self.setSelectionRectVisible(True)



import pandas as pd
from PySide6.QtCore import QAbstractItemModel, Qt


def tablemodel_to_dataframe(model: QAbstractItemModel, role: int) -> pd.DataFrame:
    data_pd = pd.DataFrame(
        columns=[
            model.headerData(c, Qt.Orientation.Horizontal)
            for c in range(model.columnCount())
        ]
    )
    for r in range(model.rowCount()):
        new_row = pd.Series(
            {
                model.headerData(c, Qt.Orientation.Horizontal): model.index(r, c).data(
                    role=role
                )
                for c in range(model.columnCount())
            }
        )
        data_pd.loc[model.headerData(r, Qt.Orientation.Vertical)] = new_row

    return data_pd



from typing import Any

import matplotlib as mpl
import numpy as np
from PySide6.QtCore import (
    QAbstractTableModel,
    QByteArray,
    QModelIndex,
    QObject,
    QPersistentModelIndex,
    QSettings,
    QSortFilterProxyModel,
    Qt,
    Signal,
    Slot,
)
from PySide6.QtGui import QBrush, QColor
from PySide6.QtSql import QSqlDatabase, QSqlQuery, QSqlRelationalTableModel
from scipy import stats

from bn_modeller.models.feature_sqltable_model import (
    FeatureSqlTableModel,
    PersistanceCheckableFeatureListProxyModel,
)
from bn_modeller.models.sample_sqltable_model import SampleSqlTableModel


class DependencyManyToManySqlTableModel(QSqlRelationalTableModel):
    table_name = "feature_dependency"
    column_source_feature_id = "source_feature_id"
    column_target_feature_id = "target_feature_id"

    def __init__(self, parent: QObject = None, db: QSqlDatabase = None):
        super().__init__(parent, db)

        query = QSqlQuery(
            f"CREATE TABLE IF NOT EXISTS {DependencyManyToManySqlTableModel.table_name} (\
                          {DependencyManyToManySqlTableModel.column_source_feature_id} INTEGER NOT NULL, \
                          {DependencyManyToManySqlTableModel.column_target_feature_id} INTEGER NOT NULL, \
                          PRIMARY KEY ({DependencyManyToManySqlTableModel.column_source_feature_id}, {DependencyManyToManySqlTableModel.column_target_feature_id}), \
                          FOREIGN KEY({DependencyManyToManySqlTableModel.column_source_feature_id}) REFERENCES {FeatureSqlTableModel.table_name}({FeatureSqlTableModel.column_id})\
                          FOREIGN KEY({DependencyManyToManySqlTableModel.column_target_feature_id}) REFERENCES {FeatureSqlTableModel.table_name}({FeatureSqlTableModel.column_id})\
                          );\
                          "
        )

        if not query.exec():
            raise RuntimeError(f"Unable to connect to DB: {query.lastError()}")
        self.setTable(DependencyManyToManySqlTableModel.table_name)


class PairTableSQLProxyModel(QAbstractTableModel):
    pairs_table_name = DependencyManyToManySqlTableModel.table_name
    samples_table_name = SampleSqlTableModel.table_name

    index_tbl_cls = FeatureSqlTableModel
    index_table_name = index_tbl_cls.table_name

    column_source_feature_id = (
        DependencyManyToManySqlTableModel.column_source_feature_id
    )
    column_target_feature_id = (
        DependencyManyToManySqlTableModel.column_target_feature_id
    )

    ValuePairsRole = Qt.ItemDataRole.UserRole + 1
    PearsonCorrRole = Qt.ItemDataRole.UserRole + 2
    SpearmanCorrRole = Qt.ItemDataRole.UserRole + 3

    def __init__(
        self,
        featureSqlTableModel: FeatureSqlTableModel,
        parent: QObject = None,
        db: QSqlDatabase = None,
    ):
        super().__init__(parent)
        self._db = db
        self._featureSqlTableModel = featureSqlTableModel

        self._cacheValuePairsRole = {}
        self._cachePearsonCorrRole = {}
        self._cacheSpearmanCorrRole = {}

        self._settings = QSettings()
        self._correlationColormap = mpl.colormaps[
            self._settings.value("depTable/colormap", "coolwarm")
        ]

    def getFeatureSqlTableModel(self):
        return self._featureSqlTableModel

    def columnCount(self, index: QModelIndex = QModelIndex()) -> int:
        return self._getFeaturesCount()

    def rowCount(self, parent: QModelIndex = QModelIndex()) -> int:
        return self._getFeaturesCount()

    def data(self, item: QModelIndex, role: int = Qt.ItemDataRole.DisplayRole) -> Any:
        if role == Qt.ItemDataRole.DisplayRole:
            return f"{self.data(item=item, role=self.PearsonCorrRole):.2f}/{self.data(item=item, role=self.SpearmanCorrRole):.2f}"
        elif role == Qt.ItemDataRole.CheckStateRole:
            return (
                Qt.CheckState.Checked
                if self._getConnectionState(item)
                else Qt.CheckState.Unchecked
            )
        elif role == self.ValuePairsRole:
            firstFeatureId, secondFeatureId = self._indexToId(index=item)
            if (firstFeatureId, secondFeatureId) not in self._cacheValuePairsRole:
                self._cacheValuePairsRole[(firstFeatureId, secondFeatureId)] = (
                    self._getFeaturePairSamples(firstFeatureId, secondFeatureId)
                )
            return self._cacheValuePairsRole[(firstFeatureId, secondFeatureId)]
        elif role == self.PearsonCorrRole:
            firstFeatureId, secondFeatureId = self._indexToId(index=item)
            if (firstFeatureId, secondFeatureId) not in self._cachePearsonCorrRole:
                values_np = self.data(item=item, role=self.ValuePairsRole)
                nas = np.logical_or(np.isnan(values_np[0]), np.isnan(values_np[1]))
                try:
                    pearsonCorr = stats.pearsonr(values_np[0, ~nas], values_np[1, ~nas])
                    self._cachePearsonCorrRole[(firstFeatureId, secondFeatureId)] = (
                        pearsonCorr.correlation
                    )
                except ValueError:
                    self._cachePearsonCorrRole[(firstFeatureId, secondFeatureId)] = (
                        np.nan
                    )
            return self._cachePearsonCorrRole[(firstFeatureId, secondFeatureId)]
        elif role == self.SpearmanCorrRole:
            firstFeatureId, secondFeatureId = self._indexToId(index=item)
            if (firstFeatureId, secondFeatureId) not in self._cacheSpearmanCorrRole:
                values_np = self.data(item=item, role=self.ValuePairsRole)
                nas = np.logical_or(np.isnan(values_np[0]), np.isnan(values_np[1]))
                spearmanrCorr = stats.spearmanr(values_np[0, ~nas], values_np[1, ~nas])
                self._cacheSpearmanCorrRole[(firstFeatureId, secondFeatureId)] = (
                    spearmanrCorr.statistic
                )
            return self._cacheSpearmanCorrRole[(firstFeatureId, secondFeatureId)]
        elif role == Qt.ItemDataRole.BackgroundRole:
            if item.column() == item.row():
                return None
            pearsonCorr = self.data(item=item, role=self.PearsonCorrRole)
            # colormap works with range [0,1], but correlation can be [-1,1]
            color = self._correlationColormap((pearsonCorr + 1) / 2, bytes=True)
            return QBrush(QColor(color[0], color[1], color[2], color[3]))
        return None

    def setData(
        self, index: QModelIndex, value: Any, role: int = Qt.ItemDataRole.DisplayRole
    ):
        if role == Qt.ItemDataRole.DisplayRole:
            return
        elif role == Qt.ItemDataRole.CheckStateRole:
            if bool(value):
                self._setConnection(index)
            else:
                self._removeConnection(index)
        self.dataChanged.emit(index, index, role)
        return True

    def headerData(
        self,
        section: int,
        orientation: Qt.Orientation,
        role: int = Qt.ItemDataRole.DisplayRole,
    ) -> Any:
        if role == Qt.ItemDataRole.DisplayRole:
            return self._getFeatureName(section)
        return None

    def flags(self, index: QModelIndex):
        if not index.isValid():
            return Qt.ItemFlag.ItemIsEnabled

        return (
            Qt.ItemFlag.ItemIsUserCheckable
            | Qt.ItemFlag.ItemIsEnabled & ~Qt.ItemFlag.ItemIsEditable
        )

    def roleNames(self) -> dict[int, QByteArray]:
        d = {}
        d[Qt.ItemDataRole.DisplayRole] = "display".encode()
        d[Qt.ItemDataRole.CheckStateRole] = "CheckState".encode()
        d[Qt.ItemDataRole.BackgroundRole] = "background".encode()

        d[self.ValuePairsRole] = "Values".encode()
        d[self.PearsonCorrRole] = "PearsonCorr".encode()
        d[self.SpearmanCorrRole] = "SpearmanCorr".encode()
        return d

    def setHeaderData(
        self,
        section: int,
        orientation: Qt.Orientation,
        value: Any,
        role: int = Qt.ItemDataRole.EditRole,
    ) -> bool:
        pass

    def canFetchMore(self, parent: QModelIndex = QModelIndex()) -> bool:
        return False

    def fetchMore(self, parent: QModelIndex = QModelIndex()) -> None:
        return

    def insertColumns(
        self, column: int, count: int, parent: QModelIndex = QModelIndex()
    ) -> bool:
        raise NotImplementedError("insertColumns is not supported")

    def removeColumns(
        self, column: int, count: int, parent: QModelIndex = QModelIndex()
    ) -> bool:
        raise NotImplementedError("removeColumns is not supported")

    def _getConnectionState(self, index: QModelIndex):
        source_id, target_id = self._indexToId(index)
        query = QSqlQuery(
            f"SELECT COUNT(*) FROM {self.pairs_table_name} WHERE {self.column_source_feature_id} = {source_id} AND {self.column_target_feature_id} = {target_id};",
            self._db,
        )
        if not query.exec():
            raise RuntimeError(
                f"Unable to retrieve row count from DB: {query.lastError()}"
            )
        query.next()
        return bool(query.value(0))

    def _setConnection(self, index: QModelIndex):
        source_id, target_id = self._indexToId(index)
        query = QSqlQuery(
            f"INSERT INTO {self.pairs_table_name}({self.column_source_feature_id}, {self.column_target_feature_id}) VALUES ({source_id}, {target_id}) ON CONFLICT DO NOTHING;",
            self._db,
        )
        if not query.exec():
            raise RuntimeError(
                f"Unable to retrieve row count from DB: {query.lastError()}"
            )

    def _removeConnection(self, index: QModelIndex):
        source_id, target_id = self._indexToId(index)
        query = QSqlQuery(
            f"DELETE FROM {self.pairs_table_name} WHERE {self.column_source_feature_id} = {source_id} AND {self.column_target_feature_id} = {target_id};",
            self._db,
        )
        if not query.exec():
            raise RuntimeError(
                f"Unable to retrieve row count from DB: {query.lastError()}"
            )

    def _indexToId(self, index: QModelIndex) -> tuple[int, int]:
        source_id = self._featureSqlTableModel.data(
            self._featureSqlTableModel.index(
                index.row(),
                self._featureSqlTableModel.fieldIndex(
                    self._featureSqlTableModel.column_id
                ),
            )
        )
        target_id = self._featureSqlTableModel.data(
            self._featureSqlTableModel.index(
                index.column(),
                self._featureSqlTableModel.fieldIndex(
                    self._featureSqlTableModel.column_id
                ),
            )
        )
        return (source_id, target_id)

    def _getFeaturesCount(self):
        query = QSqlQuery(f"SELECT COUNT(*) FROM {self.index_table_name};", self._db)
        if not query.exec():
            raise RuntimeError(
                f"Unable to retrieve row count from DB: {query.lastError()}"
            )
        query.next()
        return query.value(0)

    def _getFeatureName(self, feature_index: int):
        v = self._featureSqlTableModel.data(
            self._featureSqlTableModel.index(
                feature_index,
                self._featureSqlTableModel.fieldIndex(
                    self._featureSqlTableModel.column_name
                ),
            )
        )
        return v

    def _getFeaturePairSamples(
        self, firstFeatureId: int, secondFeatureId: int
    ) -> np.ndarray:

        query = QSqlQuery(
            f"select\
            s.{SampleSqlTableModel.column_sample_id},\
            max(case when s.{SampleSqlTableModel.column_feature_id}={firstFeatureId} then s.{SampleSqlTableModel.column_value} end) firstFeature,\
            max(case when s.{SampleSqlTableModel.column_feature_id}={secondFeatureId} then s.{SampleSqlTableModel.column_value} end) secondFeature\
            from {SampleSqlTableModel.table_name} s\
            join {FeatureSqlTableModel.table_name} f on\
            f.{FeatureSqlTableModel.column_id}=s.{SampleSqlTableModel.column_feature_id}\
            group by s.{SampleSqlTableModel.column_sample_id}\
            ",
            self._db,
        )
        if not query.exec():
            raise RuntimeError(
                f"Unable to retrieve row count from DB: {query.lastError()}"
            )
        # sampleTdValueFieldNo = query.record().indexOf(
        #     {SampleSqlTableModel.column_sample_id})
        firstFeatureFieldNo = query.record().indexOf("firstFeature")
        secondFeatureFieldNo = query.record().indexOf("secondFeature")

        values = []
        while query.next():
            # sampleTdValue: int = query.value(sampleTdValueFieldNo)
            firstFeatureValue = query.value(firstFeatureFieldNo)
            if isinstance(firstFeatureValue, str):
                firstFeatureValue = (
                    float(firstFeatureValue) if len(firstFeatureValue) > 0 else np.nan
                )
            elif isinstance(firstFeatureValue, float):
                firstFeatureValue = float(firstFeatureValue)
            else:
                raise RuntimeError(
                    f"Unexpected type for firstFeatureFieldNo({firstFeatureFieldNo}): {type(firstFeatureValue)}"
                )

            secondFeatureValue = query.value(secondFeatureFieldNo)
            if isinstance(secondFeatureValue, str):
                secondFeatureValue = (
                    float(secondFeatureValue) if len(secondFeatureValue) > 0 else np.nan
                )
            elif isinstance(secondFeatureValue, float):
                secondFeatureValue = float(secondFeatureValue)
            else:
                raise RuntimeError(
                    f"Unexpected type for secondFeatureFieldNo({secondFeatureFieldNo}): {type(secondFeatureValue)}"
                )
            values.append([firstFeatureValue, secondFeatureValue])
        return np.array(values).T


class FilterPairTableSQLProxyModel(QSortFilterProxyModel):
    filterInvalidated = Signal()

    def __init__(self, parent: QObject = None):
        super().__init__(parent)
        self.booleanSet: dict[int, bool] = {}
        self._filterModel: PersistanceCheckableFeatureListProxyModel = None

    def filterModel(self) -> PersistanceCheckableFeatureListProxyModel:
        return self._filterModel

    def setFilterModel(
        self,
        filterModel: PersistanceCheckableFeatureListProxyModel,
        filterValueColumn: int,
    ):
        if self._filterModel is not None:
            self._filterModel.dataChanged.disconnect(self.invalidateCache)
        self._filterModel = filterModel
        self._filterValueColumn = filterValueColumn
        self._filterModel.dataChanged.connect(self.invalidateCache)
        self.invalidateCache()

    def filterAcceptsRow(
        self, source_row: int, source_parent: QModelIndex | QPersistentModelIndex
    ):
        index = self._filterModel.index(
            source_row, self.filterKeyColumn(), source_parent
        )
        return self._filter_cache.get(index.data(), False)

    def filterAcceptsColumn(
        self, source_column: int, source_parent: QModelIndex | QPersistentModelIndex
    ):
        index = self._filterModel.index(
            source_column, self.filterKeyColumn(), source_parent
        )
        return self._filter_cache.get(index.data(), False)

    @Slot(QModelIndex, QModelIndex, "QList<int>")
    def invalidateCache(
        self,
        topLeft: QModelIndex = None,
        bottomRight: QModelIndex = None,
        roles: list[int] = None,
    ):
        self._filter_cache = {}
        for rowIdx in range(self._filterModel.rowCount()):
            k = self._filterModel.data(
                self._filterModel.index(rowIdx, self._filterValueColumn)
            )
            v = self._filterModel.data(
                self._filterModel.index(rowIdx, self._filterValueColumn),
                role=Qt.ItemDataRole.CheckStateRole,
            )
            self._filter_cache[k] = v == Qt.CheckState.Checked
        self.invalidateFilter()
        self.filterInvalidated.emit()



from PySide6.QtCore import (
    Property,
    QModelIndex,
    QObject,
    QPoint,
    QRect,
    QSize,
    Qt,
    Signal,
    Slot,
)
from PySide6.QtGui import QAction, QFont, QFontMetrics, QPainter
from PySide6.QtWidgets import QHeaderView, QMenu, QStyle, QStyleOptionHeader, QTableView


class RotatableHeaderView(QHeaderView):
    """Draft for Rotated header. It does not work."""

    rotate_angle_changed = Signal()

    def __init__(self, orientation: Qt.Orientation, parent: QObject = None):
        super().__init__(orientation, parent)
        self._font = QFont("helvetica", 15)
        self._metrics: QFontMetrics = QFontMetrics(self._font)
        self._descent = self._metrics.descent()
        self._margin = 10

        self._rotateAngle = 0

    def set_rotate_angle(self, angle: int):
        self._rotateAngle = angle

    def get_rotate_angle(self):
        return self._rotateAngle

    def paintSection(self, painter: QPainter, rect: QRect, index: int):
        if not rect.isValid():
            return
        opt = QStyleOptionHeader()
        self.initStyleOption(opt)

        state = QStyle.StateFlag.State_None
        if self.isEnabled():
            state |= QStyle.StateFlag.State_Enabled
        if self.window().isActiveWindow():
            state |= QStyle.StateFlag.State_Active
        if self.isSortIndicatorShown() and self.sortIndicatorSection() == index:
            opt.sortIndicator = (
                QStyleOptionHeader.SortIndicator.SortDown
                if self.sortIndicatorOrder() == Qt.SortOrder.AscendingOrder
                else QStyleOptionHeader.SortIndicator.SortUp
            )

        # setup the style options structure
        opt.rect = rect
        opt.section = index
        opt.state |= state

        opt.iconAlignment = Qt.AlignmentFlag.AlignVCenter
        opt.text = self.model().headerData(
            index, self.orientation(), Qt.ItemDataRole.DisplayRole
        )

        # @// the section position
        visual = self.visualIndex(index)
        if self.count() == 1:
            opt.position = QStyleOptionHeader.SectionPosition.OnlyOneSection
        elif visual == 0:
            opt.position = QStyleOptionHeader.SectionPosition.Beginning
        elif visual == self.count() - 1:
            opt.position = QStyleOptionHeader.SectionPosition.End
        else:
            opt.position = QStyleOptionHeader.SectionPosition.Middle

        # // the selected position

        # // draw the section

        # //store the header text
        headerText = opt.text
        # //reset the header text to no text
        opt.text = ""
        # //draw the control (unrotated!)
        self.style().drawControl(QStyle.ControlElement.CE_Header, opt, painter, self)

        painter.save()
        painter.translate(rect.x(), rect.y())
        painter.rotate(self.rotateAngle)  # // or 270
        painter.drawText(0, 0, headerText)
        painter.restore()

        # return super().paintSection(painter, rect, index)

    def sizeHint(self):
        return QSize(0, self._get_text_width() + 2 * self._margin)

    def _get_text_width(self):
        try:
            return max(
                [
                    self._metrics.horizontalAdvance(self._get_data(i))
                    for i in range(0, self.model().columnCount())
                ]
            )
        except:
            return 0

    # def _get_data(self, index):
    #     return self.model().headerData(index, self.orientation())

    rotateAngle = Property(
        str,
        fget=get_rotate_angle,
        fset=set_rotate_angle,
        notify=rotate_angle_changed,
        doc="Current rotation angle",
    )


class DependencySetupTableView(QTableView):
    def __init__(self, parent: QObject = None):
        super().__init__(parent)
        # horizontalHeaderView = RotatableHeaderView(Qt.Orientation.Horizontal)
        # horizontalHeaderView.rotateAngle = 90
        # self.setHorizontalHeader(horizontalHeaderView)
        self.initContextMenu()

    def initContextMenu(self):

        self.horizontalHeader().setContextMenuPolicy(
            Qt.ContextMenuPolicy.CustomContextMenu
        )
        self.horizontalHeader().customContextMenuRequested.connect(
            self.showHorizontalHeaderContextMenu
        )
        self.verticalHeader().setContextMenuPolicy(
            Qt.ContextMenuPolicy.CustomContextMenu
        )
        self.verticalHeader().customContextMenuRequested.connect(
            self.showVerticalHeaderContextMenu
        )

        self.headerMenu: QMenu = QMenu(self)
        selectAllAction = QAction(self.tr("Select All"), self)
        selectAllAction.triggered.connect(self.setCheckStateAll)
        self.headerMenu.addAction(selectAllAction)

        unselectAllAction = QAction(self.tr("Unselect All"), self)
        unselectAllAction.triggered.connect(self.removeCheckStateAll)
        self.headerMenu.addAction(unselectAllAction)

        swapSelectionAction = QAction(self.tr("Swap selection"), self)
        swapSelectionAction.triggered.connect(self.invertCheckStateAll)
        self.headerMenu.addAction(swapSelectionAction)

    @Slot(QPoint)
    def showHorizontalHeaderContextMenu(self, pos: QPoint):
        index: QModelIndex = self.indexAt(pos)
        self.headerMenu.popup(self.horizontalHeader().viewport().mapToGlobal(pos))

    @Slot(QPoint)
    def showVerticalHeaderContextMenu(self, pos: QPoint):
        index: QModelIndex = self.indexAt(pos)
        self.headerMenu.popup(self.verticalHeader().viewport().mapToGlobal(pos))

    @Slot()
    def setCheckStateAll(self):
        source = self.sender()
        self._setCheckStateBySender(source, True)

    @Slot()
    def removeCheckStateAll(self):
        source = self.sender()
        self._setCheckStateBySender(source, False)

    @Slot()
    def invertCheckStateAll(self):
        source = self.sender()
        if isinstance(source, QAction):
            if (
                self.horizontalHeader()
                .geometry()
                .contains(self.mapFromGlobal(self.headerMenu.pos()))
            ):
                index: QModelIndex = self.indexAt(
                    self.horizontalHeader().mapFromGlobal(self.headerMenu.pos())
                )
                print(index)
                for r in range(self.model().rowCount()):
                    i = index.siblingAtRow(r)
                    old_value = (
                        i.data(Qt.ItemDataRole.CheckStateRole) == Qt.CheckState.Checked
                    )
                    self.model().setData(
                        i, not old_value, role=Qt.ItemDataRole.CheckStateRole
                    )
            elif (
                self.verticalHeader()
                .geometry()
                .contains(self.mapFromGlobal(self.headerMenu.pos()))
            ):
                index: QModelIndex = self.indexAt(
                    self.verticalHeader().mapFromGlobal(self.headerMenu.pos())
                )
                for c in range(self.model().columnCount()):
                    i = index.siblingAtColumn(c)
                    old_value = (
                        i.data(Qt.ItemDataRole.CheckStateRole) == Qt.CheckState.Checked
                    )
                    self.model().setData(
                        i, not old_value, role=Qt.ItemDataRole.CheckStateRole
                    )

    def _setCheckStateBySender(self, source, checkState: bool):
        if isinstance(source, QAction):
            if (
                self.horizontalHeader()
                .geometry()
                .contains(self.mapFromGlobal(self.headerMenu.pos()))
            ):
                index: QModelIndex = self.indexAt(
                    self.horizontalHeader().mapFromGlobal(self.headerMenu.pos())
                )
                for r in range(self.model().rowCount()):
                    self.model().setData(
                        index.siblingAtRow(r),
                        checkState,
                        role=Qt.ItemDataRole.CheckStateRole,
                    )
            elif (
                self.verticalHeader()
                .geometry()
                .contains(self.mapFromGlobal(self.headerMenu.pos()))
            ):
                index: QModelIndex = self.indexAt(
                    self.verticalHeader().mapFromGlobal(self.headerMenu.pos())
                )
                for c in range(self.model().columnCount()):
                    self.model().setData(
                        index.siblingAtColumn(c),
                        checkState,
                        role=Qt.ItemDataRole.CheckStateRole,
                    )



import enum

from PySide6.QtCore import Property, QPoint, QRegularExpression, Qt, Signal, Slot
from PySide6.QtWidgets import (
    QFileDialog,
    QHBoxLayout,
    QLineEdit,
    QPushButton,
    QStyle,
    QWidget,
)


class FilePathWidget(QWidget):
    file_path_changed = Signal(str)

    class FilePathMode(enum.IntEnum):
        SaveFileName = (0,)
        OpenFileName = 1

    def __init__(
        self,
        caption: str,
        fileSelectionFilter: str,
        defaultDir: str = None,
        mode: FilePathMode = FilePathMode.SaveFileName,
        parent: QWidget | None = None,
    ) -> None:
        super().__init__(parent)
        self._caption: str = caption
        self._defaultDir: str = defaultDir
        self._fileSelectionFilter: str = fileSelectionFilter
        self._dialogFcn = (
            QFileDialog.getSaveFileName
            if mode == FilePathWidget.FilePathMode.SaveFileName
            else QFileDialog.getOpenFileName
        )
        self._init_ui()

    def setMode(self, mode: FilePathMode):
        self._dialogFcn = (
            QFileDialog.getSaveFileName
            if mode == FilePathWidget.FilePathMode.SaveFileName
            else QFileDialog.getOpenFileName
        )

    def _init_ui(self):
        self.root_layout = QHBoxLayout()

        self.path_edit = QLineEdit()
        self.path_edit.setPlaceholderText(self.tr("Select file path"))
        self.dialog_button = QPushButton(
            icon=self.style().standardIcon(QStyle.StandardPixmap.SP_DialogOpenButton)
        )
        self.dialog_button.clicked.connect(self.selectFile)

        self.root_layout.addWidget(self.path_edit)
        self.root_layout.addWidget(self.dialog_button)
        self.setLayout(self.root_layout)

    @Slot()
    def selectFile(self):
        fileName = self._dialogFcn(
            self, self._caption, self._defaultDir, self._fileSelectionFilter
        )
        if len(fileName[0]) > 0:
            self.file_path = fileName[0]

    def set_file_path(self, file_path: str):
        self.path_edit.setText(file_path)
        self.file_path_changed.emit(file_path)

    def get_file_path(self) -> str:
        return self.path_edit.text()

    file_path = Property(
        str,
        fget=get_file_path,
        fset=set_file_path,
        notify=file_path_changed,
        doc="Current selected path",
    )



from PySide6.QtCore import (
    QModelIndex,
    QObject,
    QPersistentModelIndex,
    QSortFilterProxyModel,
    Qt,
)


class CheckableSortFilterProxyModel(QSortFilterProxyModel):
    def __init__(self, parent: QObject = None):
        super().__init__(parent)
        self.booleanSet: dict[int, bool] = {}

    def mapFromSource(self, sourceIndex: QModelIndex | QPersistentModelIndex):
        if sourceIndex.isValid():
            return self.createIndex(sourceIndex.row(), sourceIndex.column())
        else:
            return super().mapFromSource(sourceIndex)

    def mapToSource(self, proxyIndex: QModelIndex | QPersistentModelIndex):
        if proxyIndex.isValid():
            return self.createIndex(proxyIndex.row(), proxyIndex.column())
        else:
            return super().mapToSource(proxyIndex)

    def data(self, index: QModelIndex, role: int = Qt.ItemDataRole.DisplayRole):
        if not index.isValid() or index.row() == -1:
            return None

        if role == Qt.ItemDataRole.CheckStateRole:
            return (
                Qt.CheckState.Checked
                if self.booleanSet.get(index.row(), False)
                else Qt.CheckState.Unchecked
            )
        elif role == Qt.ItemDataRole.DisplayRole:
            i = self.mapToSource(index)
            r = super().data(self.mapToSource(index), role)
            return r
        else:
            super().data(self.mapToSource(index), role)
        return super().data(index, role)

    def setData(
        self, index: QModelIndex, value, role: int = Qt.ItemDataRole.DisplayRole
    ):
        if role == Qt.ItemDataRole.CheckStateRole:
            self.booleanSet[index.row()] = bool(value)
            self.dataChanged.emit(index, index, [role])
            return True
        else:
            return super().setData(self.mapToSource(index), value, role)

    def flags(self, index: QModelIndex):
        if not index.isValid():
            return Qt.ItemFlag.ItemIsEnabled

        return (
            Qt.ItemFlag.ItemIsUserCheckable
            | Qt.ItemFlag.ItemIsEnabled
            | super().flags(self.mapToSource(index)) & ~Qt.ItemFlag.ItemIsEditable
        )

    def setSourceModel(self, sourceModel):
        self.booleanSet = {i: False for i in range(sourceModel.rowCount())}
        return super().setSourceModel(sourceModel)



import os

from PySide6.QtGui import QIcon
from PySide6.QtWidgets import (
    QDialog,
    QGridLayout,
    QLabel,
    QPlainTextEdit,
    QPushButton,
    QVBoxLayout,
)


class AboutDialog(QDialog):
    def __init__(self, parent=None):
        super().__init__(parent)
        my_icon = QIcon()
        my_icon.addFile(
            os.path.join(os.path.dirname(__file__), "..", "resources\\icon.ico")
        )
        self.setWindowIcon(my_icon)

        self.initUI()

    def initUI(self):
        layout = QVBoxLayout()
        self.setLayout(layout)

        # License text
        license_text = """
           <b>Bayesian Network Modeller</b>, <br>
            A program to represents a set of variables and their conditional dependencies via a directed acyclic graph.<br>
            Copyright (C) 2025  Digiratory <br>
            <br>
            This program is free software: you can redistribute it and/or modify<br>
            it under the terms of the GNU General Public License as published by<br>
            the Free Software Foundation, either version 3 of the License, or<br>
            (at your option) any later version.<br>
            <br>
            This program is distributed in the hope that it will be useful,<br>
            but WITHOUT ANY WARRANTY; without even the implied warranty of<br>
            MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the<br>
            GNU General Public License for more details.<br>
            <br>
            You should have received a copy of the GNU General Public License<br>
            <a href="https://www.gnu.org/licenses/">https://www.gnu.org/licenses/</a>.<br>
            <br>
            Source code can be found at <a href="https://github.com/Digiratory/bayes_model">https://github.com/Digiratory/bayes_model</a>.
            """
        license_label = QLabel(license_text)
        layout.addWidget(license_label)

        # Open-source libraries and licenses
        lib_licenses = [
            ("PySide6", "LGPL-3.0"),
            ("Qt", "LGPL-3.0"),
            ("py-banshee", "GPL-3.0"),
            ("openpyxl", "MIT"),
            ("pygraphviz", "BSD License (BSD-3-Clause)"),
            ("pyqtgraph", "BSD License (BSD-3-Clause)"),
        ]
        lib_text = "<b>3rd party libraries used in this project:</b><br>"
        for i, (lib, license) in enumerate(lib_licenses):
            lib_text += f"""
                <b>{lib}:</b> {license} <br>
                """
        libs_label = QLabel(lib_text)
        libs_label.setFixedWidth(250)
        layout.addWidget(libs_label)

        # Authors
        authors_label = QLabel("<b>Authors:</b><br}")
        layout.addWidget(authors_label)
        authors_text = """
                Aleksandr Sinitca (amsinitca@etu.ru) <a href=\"mailto:amsinitca@etu.ru\">Contact</a><br>
                Irina Shpakovskaya<br>
                Tarapata Svetlana
            """
        rich_authors_label = QLabel(authors_text)
        rich_authors_label.setOpenExternalLinks(True)
        layout.addWidget(rich_authors_label)

        # Button to close the dialog
        button_box = QPushButton("Close")
        layout.addWidget(button_box)
        button_box.clicked.connect(self.close)

    def showEvent(self, event):
        super().showEvent(event)
        self.adjustSize()



from PySide6.QtCore import (
    QIdentityProxyModel,
    QModelIndex,
    QObject,
    QSortFilterProxyModel,
    Qt,
    Signal,
    Slot,
)
from PySide6.QtSql import (
    QSqlDatabase,
    QSqlQuery,
    QSqlRelation,
    QSqlRelationalTableModel,
)


class FeatureSqlTableModel(QSqlRelationalTableModel):
    table_name = "feature"
    column_id = "id"
    column_name = "name"
    column_description = "description"
    column_is_active = "is_active"

    def __init__(self, parent: QObject = None, db: QSqlDatabase = None):
        super().__init__(parent, db)

        query = QSqlQuery(
            f"CREATE TABLE IF NOT EXISTS {FeatureSqlTableModel.table_name} (\
                          {FeatureSqlTableModel.column_id} INTEGER PRIMARY KEY AUTOINCREMENT, \
                          {FeatureSqlTableModel.column_name} TEXT, \
                          {FeatureSqlTableModel.column_description} TEXT, \
                          {FeatureSqlTableModel.column_is_active} INTEGER\
                          );\
                          "
        )

        if not query.exec():
            raise RuntimeError(f"Unable to connect to DB: {query.lastError()}")
        self.setTable(FeatureSqlTableModel.table_name)

    def setData(
        self, index: QModelIndex, value, role: int = Qt.ItemDataRole.DisplayRole
    ):
        if not index.isValid():
            return False
        if index.column() == self.fieldIndex(self.column_is_active):
            if self._setIsActiveSql(index, bool(value)):
                self.submitAll()
                self.select()
                self.dataChanged.emit(index, index, [role])
                return True
            else:
                return False
        else:
            return super().setData(index, value, role)

    def flags(self, index):
        if not index.isValid():
            return Qt.ItemFlag.ItemIsEnabled
        if index.column() == self.fieldIndex(self.column_is_active):
            return (
                Qt.ItemFlag.ItemIsUserCheckable
                | Qt.ItemFlag.ItemIsEnabled
                | Qt.ItemFlag.ItemIsEditable
            )
        else:
            return super().flags(index)

    def _setIsActiveSql(self, index: QModelIndex, value: bool):
        query = QSqlQuery()
        query.prepare(
            f"UPDATE {self.table_name} SET {self.column_is_active} = ? where {self.column_id} = ?"
        )
        query.addBindValue(value)
        query.addBindValue(
            index.siblingAtColumn(self.fieldIndex(self.column_id)).data()
        )
        return query.exec()


class PersistanceCheckableFeatureListProxyModel(QIdentityProxyModel):
    def __init__(self, parent: QObject = None):
        super().__init__(parent)

    def setSourceModel(self, sourceModel: FeatureSqlTableModel):
        return super().setSourceModel(sourceModel)

    def sourceModel(self) -> FeatureSqlTableModel:
        return super().sourceModel()

    def data(self, index: QModelIndex, role: int = Qt.ItemDataRole.DisplayRole):
        sourceIndex: QModelIndex = self.mapToSource(index)
        if role == Qt.ItemDataRole.CheckStateRole:
            is_checked = sourceIndex.siblingAtColumn(
                self.sourceModel().fieldIndex(self.sourceModel().column_is_active)
            ).data()
            return Qt.CheckState.Checked if is_checked else Qt.CheckState.Unchecked
        return super().data(index, role)

    def setData(
        self, index: QModelIndex, value, role: int = Qt.ItemDataRole.DisplayRole
    ):
        sourceIndex: QModelIndex = self.mapToSource(index)
        if role == Qt.ItemDataRole.CheckStateRole:
            flagIndex = sourceIndex.siblingAtColumn(
                self.sourceModel().fieldIndex(self.sourceModel().column_is_active)
            )
            if super().setData(flagIndex, bool(value)):
                self.dataChanged.emit(index, index, [role])
                return True
            else:
                return False
        else:
            return super().setData(sourceIndex, value, role)

    def flags(self, index: QModelIndex):
        if not index.isValid():
            return Qt.ItemFlag.ItemIsEnabled
        return (
            Qt.ItemFlag.ItemIsUserCheckable
            | Qt.ItemFlag.ItemIsEnabled
            | super().flags(self.mapToSource(index)) & ~Qt.ItemFlag.ItemIsEditable
        )



from typing import Callable

import numpy as np
import pandas as pd
from PySide6.QtCore import (
    QModelIndex,
    QObject,
    QPersistentModelIndex,
    QSortFilterProxyModel,
    Qt,
)
from PySide6.QtSql import QSqlIndex, QSqlTableModel

from bn_modeller.models.feature_sqltable_model import FeatureSqlTableModel
from bn_modeller.models.sample_sqltable_model import SampleSqlTableModel


class SampleFilterProxyModel(QSortFilterProxyModel):
    def __init__(self, parent: QObject | None = None):
        super().__init__(parent)
        self._sampleId = None

    def enableFilter(self, sampleId):
        self._sampleId = sampleId
        self.invalidateFilter()

    def disableFilter(self):
        self._sampleId = None
        self.invalidateFilter()

    def filterAcceptsRow(
        self, source_row: int, source_parent: QModelIndex | QPersistentModelIndex
    ):
        index = self.sourceModel().index(
            source_row,
            self.sourceModel().record().indexOf(SampleSqlTableModel.column_sample_id),
            source_parent,
        )
        if self._sampleId is not None:
            data = self.sourceModel().data(index)
            return data == self._sampleId

        return super().filterAcceptsRow(source_row, source_parent)


def to_numeric_with_callback(data: str | None, callback: Callable[[str], float]):
    """Convert a string to numeric value with a callback function for user correction.
    Args:
        data (str | None): The string to be converted.
        callback (Callable[[str], float]): A function that takes a string as input and returns a numeric value.
    Returns:
        float | None: A numeric value or None if an error occurred during conversion.
    """
    try:
        return pd.to_numeric(data, errors="raise")
    except ValueError as e:
        return callback(data)


def sheetfile_to_dataframe(
    sheet_file_path: str,
    transposed_csv: bool,
    skip_rows: int = 0,
    skip_cols: int = 0,
    value_fixer_callback: Callable[[str], float] | None = None,
):
    """Convert a CSV file to a Pandas DataFrame.
    Args:
        csv_file_path (str): The path to the CSV file.
        transposed_csv (bool): True if the CSV file is transposed.
        skip_rows (int): The number of rows to skip at the beginning of the CSV file.
        skip_cols (int): The number of columns to skip at the beginning of the CSV file.
    Returns:
        pandas.DataFrame: A Pandas DataFrame containing the CSV file's contents.
    """
    if sheet_file_path.endswith(".xlsx") or sheet_file_path.endswith(".xls"):
        data_pd = pd.read_excel(
            sheet_file_path, index_col=skip_cols, skiprows=skip_rows
        )
    elif sheet_file_path.endswith(".csv"):
        data_pd = pd.read_csv(sheet_file_path, index_col=skip_cols, skiprows=skip_rows)
    else:
        raise ValueError(f"Unsupported file format for {sheet_file_path}")

    if skip_cols > 0:
        # Delete first skip_rows rows
        data_pd = data_pd.drop(data_pd.columns[:skip_cols], axis=1)
    if transposed_csv:
        # Transpose dataframe
        data_pd = data_pd.T

    # Curate values in the dataframe to be numeric
    # TODO: ask user for correct errors
    # data_pd = data_pd.map(pd.to_numeric, errors='coerce')

    def _value_fixer(x: str):
        x_fixed = np.nan
        try:
            x_fixed = pd.to_numeric(x, errors="raise")
        except ValueError:
            if value_fixer_callback is not None:
                x_fixed = value_fixer_callback(x)
        return x_fixed

    data_pd = data_pd.map(_value_fixer)

    data_pd.index = data_pd.index.map(pd.to_numeric)
    data_pd.index = data_pd.index.astype(int)
    return data_pd


def add_values_from_csv(
    csv_file_path: str,
    transposed_csv: bool,
    featureSqlTableModel: FeatureSqlTableModel,
    sampleSqlTableModel: SampleSqlTableModel,
    skip_rows: int = 0,
    skip_cols: int = 0,
    value_fixer_callback=None,
):
    """Add values from a CSV file to the database.

    Args:
        csv_file_path (str): The path to the CSV file.
        transposed_csv (bool): True if the CSV file is transposed.
        featureSqlTableModel (FeatureSqlTableModel): The target data model for the features.
        sampleSqlTableModel (SampleSqlTableModel): The target data model for the samples.
        skip_rows (int): The number of rows to skip at the beginning of the CSV file. Default is 0.
        skip_cols (int): The number of columns to skip at the beginning of the CSV file. Default is 0.
    """
    data_pd = sheetfile_to_dataframe(
        csv_file_path,
        transposed_csv=transposed_csv,
        skip_rows=skip_rows,
        skip_cols=skip_cols,
        value_fixer_callback=value_fixer_callback,
    )

    feature_proxy = QSortFilterProxyModel()
    feature_proxy.setSourceModel(featureSqlTableModel)
    feature_proxy.setFilterKeyColumn(
        featureSqlTableModel.fieldIndex(featureSqlTableModel.column_name)
    )

    new_features = []
    for col_candidate in data_pd.columns:
        feature_proxy.setFilterFixedString(col_candidate)
        if feature_proxy.rowCount() == 0:
            new_features.append(col_candidate)

    for new_feature in new_features:
        rowRecord = featureSqlTableModel.record()
        rowRecord.remove(rowRecord.indexOf(FeatureSqlTableModel.column_id))
        rowRecord.setValue(FeatureSqlTableModel.column_name, new_feature)
        rowRecord.setValue(FeatureSqlTableModel.column_is_active, 1)
        rowRecord.setValue(FeatureSqlTableModel.column_description, "")
        featureSqlTableModel.insertRecord(-1, rowRecord)
    featureSqlTableModel.submitAll()

    # sample_proxy = SampleFilterProxyModel()
    # sample_proxy.setSourceModel(sampleSqlTableModel)
    # new_samples = []
    # for sample_candidate in data_pd.index:
    #     sample_proxy.enableFilter(sample_candidate)
    #     if sample_proxy.rowCount() == 0:
    #         new_samples.append(sample_candidate)
    # Logic with searching dublication is extremelly slow, ignore DB update as a temporal solution.
    # TODO: reimplement with raw sql query.
    if sampleSqlTableModel.rowCount() > 0:
        return

    # for new_sample in new_samples:
    # row_pd = data_pd.loc[new_sample]
    for index, row_pd in data_pd.iterrows():
        for col in data_pd.columns:
            if not np.isnan(row_pd[col]):
                rowRecord = sampleSqlTableModel.record()
                rowRecord.remove(rowRecord.indexOf(SampleSqlTableModel.column_id))

                feature_proxy.setFilterFixedString(col)

                rowRecord.setValue(
                    SampleSqlTableModel.column_feature_id,
                    feature_proxy.index(
                        0,
                        featureSqlTableModel.fieldIndex(FeatureSqlTableModel.column_id),
                    ).data(),
                )
                rowRecord.setValue(SampleSqlTableModel.column_sample_id, int(index))
                rowRecord.setValue(SampleSqlTableModel.column_value, float(row_pd[col]))
                sampleSqlTableModel.insertRecord(-1, rowRecord)
    sampleSqlTableModel.submitAll()
    # for index, row in data_pd.iterrows():
